/* 2007sep02 - Alain Brobecker */
/* 0: empty, 1:black, 2:white, so 3-player gives value for opponent */
/*!!! unsigned int must be 32 bits long */
/*!!! To remove symmetries make ComparePos always returns -1*/
/*    and remove the symmetries in PackPosition */
/*ply;games;different positions;won games;directly won pos;draw pos;total nb links
0;1;1;0;0;1;0
1;9;3;0;0;3;3
2;72;12;0;0;5;15
3;504;38;0;0;21;81
4;3024;108;0;0;18;279
5;15120;174;1440;21;35;771
6;54720;204;5328;21;27;1355
7;148176;153;47952;58;27;1881
8;200448;57;72576;23;11;2062
9;127872;15;81792;12;3;2096
Total number of positions: 1+3+12+38+...=764
number of drawn games: 127872-81792=46080
number of drawn positions: 15-12=3
Total number of games: 1440+5328+...+81792+46080=255168
Drawn positions are
  oxo   oxo   oxo
  oxx   oox   oxo
  xoo   xox   xox
Six positions allow to recover what was the last move
  xoo   xxo   xxo   xox   oxo   xox
  xox   xxo   xox   xox   xox   ooo
  oox   ooo   ooo   ooo   oxo   xox */

#include <stdio.h>
#include <stdlib.h>
#define TRUE 1
#define FALSE 0

#define texW 17.0 /* for TeX output */
#define texH 9.0
#define BoardN 3 /* also change (un)PackPosition functions, PkBoardSize, WonTest */
#define PkBoardSize 1 /* number of uints for a packed board */
#define StartTexPly 0
#define EndTexPly 1
#define MaxPly 9
#define MaxPosNb 20000
#define MaxLinkNb 20000

/**** Global variables ****/
int LegalMove;
unsigned char PlayerToMove;
unsigned int Pos[MaxPosNb*PkBoardSize],Occurences[MaxPosNb],Status[MaxPosNb];
unsigned int PkPos[PkBoardSize];
int FirstPos[MaxPly+2],LastPos,Ply,SrcPos;
float PosX[MaxPosNb],PosY[MaxPosNb];
unsigned int LinkS[MaxLinkNb],LinkE[MaxLinkNb];
int FirstLink[MaxPly+2],LastLink;

/**** Print a BoardN*BoardN board ****/
void PrintBoard(A) unsigned char A[BoardN][BoardN]; {
  int x,y;
  for(y=0;y<BoardN;y++) {
    for(x=0;x<BoardN;x++) {
      switch(A[x][y]) {
        case 0:printf("."); break;
        case 1:printf("1"); break;
        case 2:printf("2"); break;
        default:printf("-");
      }
    }
    printf("\n");
  }
}

/*!!! for debugging */
void PrintPk(A) unsigned int A; {
  int x,c; c=A;
  for(x=0;x<9;x++) {
    switch(c & 3) {
      case 0:printf("."); break;
      case 1:printf("X"); break;
      case 2:printf("O"); break;
      default:printf("-");
    }
    c=c>>2;
  }
}

/**** Copy a BoardN*BoardN board to another ****/
void CopyAToB(A,B) unsigned char A[BoardN][BoardN], B[BoardN][BoardN]; {
  int x,y;
  for(y=0;y<BoardN;y++) { for(x=0;x<BoardN;x++) { B[x][y]=A[x][y]; } }
}

/**** Packs central part of a BoardN^2 into PkBoardSize uint ****/
/* Due to the 8 similar position resulting from symmetries, we must */
/* pack all 8 positions and return only the one with lowest value.  */
/* Two positions with same value are symmetric.             abc     */
/* For 3*3 tic-tac-toe, we use the notation aside           hid     */
/* & tricks to compute symmetries more efficiently.         gfe     */
void PackPosition(A,B) unsigned char A[BoardN][BoardN]; unsigned int B[]; {
  unsigned int t0,t1,t2,t3,t4;
  t0=(A[0][0]<<8)+A[2][2];        /* t0=a...e */
  t1=(A[1][0]<<8)+A[1][2];        /* t1=b...f */
  t2=(A[2][0]<<8)+A[0][2];        /* t2=c...g */
  t3=(A[2][1]<<8)+A[0][1];        /* t3=d...h */
  t4=(t0<<6)+(t1<<4)+(t2<<2)+t3;  /* t4=abcdefgh */
  t0=(t3<<6)+(t2<<4)+(t1<<2)+t0;  /* t0=dcbahgfe, not exactly correct */
  t0=((t0<<2)+A[2][1]) & 65535;   /* t0=cbahgfed */
  /* t0 will contain min of the 4 rotated version of cbahgfed */
  t1=t0;  
  t1=(((t1<<16)+t1)>>4) & 65535;  /* rotate t1=dcbahgfe */
  if(t1<t0) { t0=t1; }            /* set t0 to it if lower */
  t1=(((t1<<16)+t1)>>4) & 65535; if(t1<t0) { t0=t1; } /* again */
  t1=(((t1<<16)+t1)>>4) & 65535; if(t1<t0) { t0=t1; } /* and again */
  /* t4 will contain min of the 4 rotated version of abcdefgh */
  t1=t4;  
  t1=(((t1<<16)+t1)>>4) & 65535;  /* rotate t1=abcdefgh */
  if(t1<t4) { t4=t1; }            /* set t4 to it if lower */
  t1=(((t1<<16)+t1)>>4) & 65535; if(t1<t4) { t4=t1; } /* again */
  t1=(((t1<<16)+t1)>>4) & 65535; if(t1<t4) { t4=t1; } /* and again */
  /* now take min of all 8 positions, add i and quit */
  if(t4<t0) { t0=t4; }
  t0=(t0<<2)+A[1][1];
  B[0]=t0;
}

/**** Unpacks position into central part of a BoardN^2 board ****/
/* abc The 3*3 board is packed as abcdefghi with the notation aside */
/* hid */
/* gfe */
void UnpackPosition(A,B) unsigned int A[]; unsigned char B[BoardN][BoardN]; {
  unsigned int c;
  c=A[0];
  B[1][1]=c & 3; c=c>>2; B[0][1]=c & 3; c=c>>2; B[0][2]=c & 3; c=c>>2;
  B[1][2]=c & 3; c=c>>2; B[2][2]=c & 3; c=c>>2; B[2][1]=c & 3; c=c>>2;
  B[2][0]=c & 3; c=c>>2; B[1][0]=c & 3; c=c>>2; B[0][0]=c & 3;
}

/**** Compare two packed positions ****/
/* returns 0 if A==B, else returns -1 */
int ComparePositions(A,B) unsigned int A[],B[]; {
  int cmp;
  if(A[0]==B[0]) { cmp=0; } else { cmp=-1; }
  return(cmp);
}

/**** Search PkPos between FirstPos[Ply] & LastPos, if ****/
/**** found we add it at the end of the unsorted list. ****/
void InsertPkPos() {
  int PosFound,CurrentPos,c,d,LinkFound;
  /* test if the position already exists using dichotomy */
  PosFound=FALSE;
  CurrentPos=FirstPos[Ply];
  while((CurrentPos<LastPos) && (PosFound==FALSE)) {
    if(ComparePositions(PkPos,Pos+CurrentPos*PkBoardSize)==0) {
      PosFound=TRUE;
      Occurences[CurrentPos]+=Occurences[SrcPos];
      /* check if link already exist */
      LinkFound=FALSE;
      for(d=0;d<LastLink;d++) {
        if((LinkS[d]==SrcPos) && (LinkE[d]==CurrentPos)) {
          LinkFound=TRUE;
        }
      }
      /* link not found, so create it */
      if(LinkFound==FALSE) {
        LinkS[LastLink]=SrcPos;
        LinkE[LastLink]=CurrentPos;
        LastLink++;
      }
    }
    CurrentPos++;
  }
  if(PosFound==FALSE)  {
    for(c=0;c<PkBoardSize;c++) { Pos[LastPos*PkBoardSize+c]=PkPos[c]; }
    Occurences[LastPos]=Occurences[SrcPos];
    LinkS[LastLink]=SrcPos;
    LinkE[LastLink]=LastPos;
    LastLink++;
    LastPos++;
  }
}

/********************************************************************** MAIN */
int main() {
  unsigned char NewBoard[BoardN][BoardN];
  int x,y,c,GameWon,NbGames,NbGamesWon,NbPosWon,NbPos;
  int NbDrawPos[MaxPly+1],MaxNbDrawPos,CurrentStatus;
  int MaxNbPos=0;
  float u,psx,psy,psx2,psy2;
  /**** Initial position ****/  
  unsigned char Board[BoardN][BoardN]={ {0,0,0},{0,0,0},{0,0,0} };

  setbuf(stdout, NULL); /* no buffering for output! */
  PackPosition(Board,Pos); /* !!! PrintBoard(Board); */
  FirstPos[0]=0;
  FirstPos[1]=1;
  Occurences[0]=1;
  FirstLink[1]=0;
  for(c=0;c<MaxNbPos;c++) { Status[c]=0; }
/*!!!  printf("!!! won games & pos are for previous ply !!!\n");*/
/*!!!  printf("ply;games;different positions;won games;won pos;total nb links\n");*/
  PlayerToMove=1;
  for(Ply=1;Ply<MaxPly+2;Ply++) {
/* printf("*** PLY %d ***\n",Ply-1); */
    NbGames=0;
    NbGamesWon=0;
    NbPosWon=0;
    LastPos=FirstPos[Ply];
    LastLink=FirstLink[Ply];
    for(SrcPos=FirstPos[Ply-1];SrcPos<FirstPos[Ply];SrcPos++) {
      UnpackPosition(Pos+(SrcPos*PkBoardSize),Board); /* PrintBoard(Board); */
      /* check if game is already finished */
      GameWon=FALSE;
      for(x=0;(x<3) && (GameWon==FALSE);x++) { /* test lines */
        c=Board[x][0]+Board[x][1]+Board[x][2];
        if((c==6) || ((c==3) && (Board[x][0]==Board[x][1]))) { GameWon=TRUE; }
      }
      for(y=0;(y<3) && (GameWon==FALSE);y++) { /* test columns */
        c=Board[0][y]+Board[1][y]+Board[2][y];
        if((c==6) || ((c==3) && (Board[0][y]==Board[1][y]))) { GameWon=TRUE; }
      }
      if(GameWon==FALSE) {
        c=Board[0][0]+Board[1][1]+Board[2][2]; /* test 1st diag */
        if((c==6) || ((c==3) && (Board[0][0]==Board[1][1]))) {
          GameWon=TRUE;
        } else {
          c=Board[0][2]+Board[1][1]+Board[2][0]; /* test 2nd diag */
          if((c==6) || ((c==3) && (Board[0][2]==Board[1][1]))) { GameWon=TRUE; }
        }
      }
      /* if game was won (at precedent turn) count it, else game continues */
      if(GameWon==TRUE) {
        NbPosWon++;
        NbGamesWon+=Occurences[SrcPos]; /* !!! PrintBoard(Board);*/
        Status[SrcPos]=3-PlayerToMove; /* Player to move for "Qui Perd Gagne" */
      } else if(Ply!=MaxPly+1) {
        for(y=0;y<BoardN;y++) {
          for(x=0;x<BoardN;x++) {
            if(Board[x][y]==0) {
              NbGames+=Occurences[SrcPos];
              CopyAToB(Board,NewBoard);
              NewBoard[x][y]=PlayerToMove;
              PackPosition(NewBoard,PkPos); /* !!! PrintPk(PkPos); */
              InsertPkPos(); /* links set during this */
            }
          } /*next x*/
        } /*next y*/
      }
    } /* next SrcPos */
    FirstPos[Ply+1]=LastPos;
    FirstLink[Ply+1]=LastLink;
    NbPos=LastPos-FirstPos[Ply];
/*    printf("%d;%d;%d;%d;%d;%d\n",Ply,NbGames,NbPos,NbGamesWon,NbPosWon,FirstLink[Ply+1]-FirstLink[Ply]);*/
    PlayerToMove=3-PlayerToMove;
    if((Ply>=StartTexPly) && (Ply<=EndTexPly) && (NbPos>MaxNbPos)) { MaxNbPos=NbPos; }
  } /*next Ply*/
  /* Compute (up to bottom) the status of every undetermined position */
  /* 0==draw, 1==win for player 1, 2==win for player 2 */
  PlayerToMove=1;
  NbDrawPos[MaxPly]=3;
  MaxNbDrawPos=3; /* 3 draw games at plys 9 */
  for(Ply=MaxPly-1;Ply>=0;Ply--) {
    NbDrawPos[Ply]=0;
    for(c=FirstPos[Ply];c<FirstPos[Ply+1];c++) {
      if(Status[c]==0) { /* undetermined? */
        CurrentStatus=3-PlayerToMove;
        x=FirstLink[Ply+1];
        while((CurrentStatus!=PlayerToMove) && (x<FirstLink[Ply+2])) {
          if(LinkS[x]==c) { /* this link concerns current pos? */
            if(Status[LinkE[x]]==PlayerToMove) {
              CurrentStatus=PlayerToMove; /* win if there's one winning move! */
            } else if(Status[LinkE[x]]<CurrentStatus) {
              CurrentStatus=Status[LinkE[x]]; /* else minimise opponent's value */
            }
          }
          x++;
        } /* next link */
        Status[c]=CurrentStatus;
        /*!!! UnpackPosition(Pos+c*PkBoardSize,Board); PrintBoard(Board); printf(" %d\n\n",CurrentStatus); */
        if(CurrentStatus==0) { NbDrawPos[Ply]++; }
      }
    } /* next pos */
/*    printf("ply %d: %d draw pos\n",Ply,NbDrawPos); */
    if(NbDrawPos[Ply]>MaxNbDrawPos) {
      if((Ply>=StartTexPly) && (Ply<=EndTexPly)) {
        MaxNbDrawPos=NbDrawPos[Ply];
      }
    }
    PlayerToMove=3-PlayerToMove;
  } /* next plys */
  
  /* TeX output */
  printf("\\documentclass[a4paper]{minimal}");
  printf("\\usepackage{amssymb}\n");
  printf("\\usepackage{lambda,ifthen,calc}\n");
  printf("\\usepackage{pstricks}\n");
  printf("\\setlength{\\textwidth}{18cm}\n");
  printf("\\setlength{\\marginparwidth}{0cm}\n");
  printf("\\setlength{\\oddsidemargin}{-1in+1.5cm}\n");
  printf("\\setlength{\\evensidemargin}{-1in+1.5cm}\n");
  printf("\\setlength{\\textheight}{27.7cm}\n");
  printf("\\setlength{\\topmargin}{0cm}\n");
  printf("\\setlength{\\headheight}{0cm}\n");
  printf("\\setlength{\\headsep}{0cm}\n");
  printf("\\setlength{\\footskip}{0cm}\n");
  printf("\\setlength{\\voffset}{-1in+1cm}\n");
  printf("\\pagestyle{empty}\n");
  printf("\\begin{document}\n");
  printf("\\begin{pspicture}(%f,%f)\n",texW,texH);
  u=texW/((float) (MaxNbPos+1)*4);
  psy=1.5*u+0.1;
  for(Ply=StartTexPly;Ply<EndTexPly+1;Ply++) {
    NbPos=FirstPos[Ply+1]-FirstPos[Ply];
    psx=0;
    for(c=FirstPos[Ply];c<FirstPos[Ply+1];c++) {
      psx+=texW/((float) NbPos+1);
      PosX[c]=psx;
      PosY[c]=psy;
      UnpackPosition(Pos+c*PkBoardSize,Board);
      printf("\\rput(%f,%f) { ",psx,psy);
      printf("\\psline[linewidth=%f](%f,%f)(%f,%f) ",0.05*u,-0.5*u,-1.5*u,-0.5*u,1.5*u);
      printf("\\psline[linewidth=%f](%f,%f)(%f,%f) ",0.05*u,0.5*u,-1.5*u,0.5*u,1.5*u);
      printf("\\psline[linewidth=%f](%f,%f)(%f,%f) ",0.05*u,-1.5*u,-0.5*u,1.5*u,-0.5*u);
      printf("\\psline[linewidth=%f](%f,%f)(%f,%f) ",0.05*u,-1.5*u,0.5*u,1.5*u,0.5*u);
      for(y=0;y<BoardN;y++) {
        for(x=0;x<BoardN;x++) {
          switch(Board[x][y]) {
            case 1:printf("\\pscircle[linewidth=%f](%f,%f){%f} ",0.1*u,((float) x-1)*u,((float) y-1)*u,0.4*u); break;
            case 2:printf("\\pscircle*[linewidth=%f](%f,%f){%f} ",0.1*u,((float) x-1)*u,((float) y-1)*u,0.4*u); break;
          }
        }
      }
      printf("}\n");
    }
    psy+=(texH-3*u-0.2)/((float) EndTexPly-StartTexPly);
  }
  /* draw links between diagrams */
  for(c=FirstLink[StartTexPly+1];c<FirstLink[EndTexPly+1];c++) {
    psx=PosX[LinkS[c]];
    psy=PosY[LinkS[c]]+2*u;
    psx2=PosX[LinkE[c]];
    psy2=PosY[LinkE[c]]-2*u;
    printf("\\psline[linewidth=%f]{->}(%f,%f)(%f,%f) ",0.05*u,psx,psy,psx2,psy2);
    printf("\\rput(%f,%f) { %d }\n",(psx+psx2)/2,(psy+psy2)/2,Occurences[LinkE[c]]/Occurences[LinkS[c]]);
  }
  printf("\\psframe(0,0)(%f,%f)\n",texW,texH);
  printf("\\end{pspicture}\n");
  printf("\\end{document}\n");

  return(0);
}
